@@ -31,7 +31,6 @@
 #include "androidconstants.h"
 #include "androidmanager.h"
 #include "androidsdkmanager.h"
-#include "androidqtsupport.h"
 #include "certificatesmodel.h"
 
 #include "javaparser.h"
@@ -40,20 +39,26 @@
 #include <coreplugin/icore.h>
 
 #include <projectexplorer/buildconfiguration.h>
+#include <projectexplorer/buildsteplist.h>
+#include <projectexplorer/processparameters.h>
 #include <projectexplorer/project.h>
 #include <projectexplorer/projectexplorerconstants.h>
+#include <projectexplorer/projectnodes.h>
+#include <projectexplorer/runconfiguration.h>
 #include <projectexplorer/target.h>
 
 #include <qtsupport/qtkitinformation.h>
 
 #include <utils/algorithm.h>
+#include <utils/infolabel.h>
+#include <utils/qtcprocess.h>
 #include <utils/synchronousprocess.h>
-#include <utils/utilsicons.h>
-
-#include <qmakeprojectmanager/qmakeprojectmanagerconstants.h>
 
+#include <QDateTime>
 #include <QDialogButtonBox>
 #include <QHBoxLayout>
+#include <QJsonDocument>
+#include <QJsonObject>
 #include <QLabel>
 #include <QLineEdit>
 #include <QLoggingCategory>
@@ -65,16 +70,16 @@
 #include <memory>
 
 using namespace ProjectExplorer;
+using namespace QtSupport;
+using namespace Utils;
 using namespace Android::Internal;
 
 namespace {
-Q_LOGGING_CATEGORY(buildapkstepLog, "qtc.android.build.androidbuildapkstep", QtWarningMsg)
+static Q_LOGGING_CATEGORY(buildapkstepLog, "qtc.android.build.androidbuildapkstep", QtWarningMsg)
 }
 
 namespace Android {
 
-const Core::Id ANDROID_BUILD_APK_ID("QmakeProjectManager.AndroidBuildApkStep");
-
 const QVersionNumber gradleScriptRevokedSdkVersion(25, 3, 0);
 const char KeystoreLocationKey[] = "KeystoreLocation";
 const char BuildTargetSdkKey[] = "BuildTargetSdk";
@@ -87,12 +92,10 @@ static void setupProcessParameters(ProcessParameters *pp,
                                    const QString &command)
 {
     pp->setMacroExpander(bc->macroExpander());
-    pp->setWorkingDirectory(bc->buildDirectory().toString());
+    pp->setWorkingDirectory(bc->buildDirectory());
     Utils::Environment env = bc->environment();
     pp->setEnvironment(env);
-    pp->setCommand(command);
-    pp->setArguments(Utils::QtcProcess::joinArgs(arguments));
-    pp->resolveAll();
+    pp->setCommandLine({command, arguments});
 }
 
 class PasswordInputDialog : public QDialog
@@ -116,22 +119,39 @@ private:
     std::function<bool (const QString &)> verifyCallback = [](const QString &) { return true; };
     QLabel *inputContextlabel = new QLabel(this);
     QLineEdit *inputEdit = new QLineEdit(this);
-    QLabel *warningIcon = new QLabel(this);
-    QLabel *warningLabel = new QLabel(this);
+    Utils::InfoLabel *warningLabel = new Utils::InfoLabel(tr("Incorrect password."),
+                                                          Utils::InfoLabel::Warning, this);
     QDialogButtonBox *buttonBox = new QDialogButtonBox(QDialogButtonBox::Ok | QDialogButtonBox::Cancel,
                                                        this);
 };
 
-AndroidBuildApkStep::AndroidBuildApkStep(BuildStepList *parent)
-    : AbstractProcessStep(parent, ANDROID_BUILD_APK_ID),
+static FilePath aabPath(const BuildConfiguration *bc)
+{
+    if (!bc)
+        return {};
+
+    QString buildType;
+    if (bc->buildType() == BuildConfiguration::Release)
+        buildType = "release";
+    else
+        buildType = "debug";
+
+    return bc->buildDirectory()
+            .pathAppended(Constants::ANDROID_BUILDDIRECTORY)
+            .pathAppended(QString("build/outputs/bundle/%1/android-build-%1.aab").arg(buildType));
+}
+
+AndroidBuildApkStep::AndroidBuildApkStep(BuildStepList *parent, Core::Id id)
+    : AbstractProcessStep(parent, id),
       m_buildTargetSdk(AndroidConfig::apiLevelNameFor(AndroidConfigurations::
                                          sdkManager()->latestAndroidSdkPlatform()))
 {
     //: AndroidBuildApkStep default display name
     setDefaultDisplayName(tr("Build Android APK"));
+    setImmutable(true);
 }
 
-bool AndroidBuildApkStep::init(QList<const BuildStep *> &earlierSteps)
+bool AndroidBuildApkStep::init()
 {
     ProjectExplorer::BuildConfiguration *bc = buildConfiguration();
 
@@ -148,13 +168,14 @@ bool AndroidBuildApkStep::init(QList<const BuildStep *> &earlierSteps)
                            OutputFormat::ErrorMessage);
     }
 
-    QtSupport::BaseQtVersion *version = QtSupport::QtKitInformation::qtVersion(target()->kit());
+    QtSupport::BaseQtVersion *version = QtSupport::QtKitAspect::qtVersion(target()->kit());
     if (!version)
         return false;
 
     const QVersionNumber sdkToolsVersion = AndroidConfigurations::currentConfig().sdkToolsVersion();
-    if (sdkToolsVersion >= gradleScriptRevokedSdkVersion) {
-        if (!version->sourcePath().appendPath("src/3rdparty/gradle").exists()) {
+    if (sdkToolsVersion >= gradleScriptRevokedSdkVersion
+        || AndroidConfigurations::currentConfig().isCmdlineSdkToolsInstalled()) {
+        if (!version->sourcePath().pathAppended("src/3rdparty/gradle").exists()) {
             emit addOutput(tr("The installed SDK tools version (%1) does not include Gradle "
                               "scripts. The minimum Qt version required for Gradle build to work "
                               "is %2").arg(sdkToolsVersion.toString()).arg("5.9.0/5.6.3"),
@@ -177,47 +198,55 @@ bool AndroidBuildApkStep::init(QList<const BuildStep *> &earlierSteps)
 
     auto parser = new JavaParser;
     parser->setProjectFileList(Utils::transform(target()->project()->files(ProjectExplorer::Project::AllFiles),
-                                                &Utils::FileName::toString));
+                                                &Utils::FilePath::toString));
 
-    AndroidQtSupport *qtSupport = AndroidManager::androidQtSupport(target());
-    QFileInfo sourceDirInfo(qtSupport->targetData(Constants::AndroidPackageSourceDir, target()).toString());
-    parser->setSourceDirectory(Utils::FileName::fromString(sourceDirInfo.canonicalFilePath()));
-    parser->setBuildDirectory(Utils::FileName::fromString(bc->buildDirectory().appendPath(Constants::ANDROID_BUILDDIRECTORY).toString()));
+    RunConfiguration *rc = target()->activeRunConfiguration();
+    const QString buildKey = rc ? rc->buildKey() : QString();
+    const ProjectNode *node = rc ? target()->project()->findNodeForBuildKey(buildKey) : nullptr;
+
+    QString sourceDirName;
+    if (node)
+        sourceDirName = node->data(Constants::AndroidPackageSourceDir).toString();
+
+    QFileInfo sourceDirInfo(sourceDirName);
+    parser->setSourceDirectory(Utils::FilePath::fromString(sourceDirInfo.canonicalFilePath()));
+    parser->setBuildDirectory(bc->buildDirectory().pathAppended(Constants::ANDROID_BUILDDIRECTORY));
     setOutputParser(parser);
 
     m_openPackageLocationForRun = m_openPackageLocation;
-    m_apkPath = AndroidManager::apkPath(target()).toString();
-    qCDebug(buildapkstepLog) << "APK path:" << m_apkPath;
+    m_packagePath = m_buildAAB ? aabPath(buildConfiguration()).toString()
+                               : AndroidManager::apkPath(target()).toString();
+    qCDebug(buildapkstepLog) << "Package path:" << m_packagePath;
 
-    if (!AbstractProcessStep::init(earlierSteps))
+    if (!AbstractProcessStep::init())
         return false;
 
-    QString command = version->qmakeProperty("QT_HOST_BINS");
+    QString command = version->hostBinPath().toString();
     if (!command.endsWith('/'))
         command += '/';
-    command += "androiddeployqt";
-    if (Utils::HostOsInfo::isWindowsHost())
-        command += ".exe";
+    command += Utils::HostOsInfo::withExecutableSuffix("androiddeployqt");
 
-    QString outputDir = bc->buildDirectory().appendPath(Constants::ANDROID_BUILDDIRECTORY).toString();
+    QString outputDir = bc->buildDirectory().pathAppended(Constants::ANDROID_BUILDDIRECTORY).toString();
 
-    QString inputFile = AndroidManager::androidQtSupport(target())
-            ->targetData(Constants::AndroidDeploySettingsFile, target()).toString();
-    if (inputFile.isEmpty()) {
+    if (node)
+        m_inputFile = node->data(Constants::AndroidDeploySettingsFile).toString();
+
+    if (m_inputFile.isEmpty()) {
+        qCDebug(buildapkstepLog) << "no input file" << rc << node << buildKey;
         m_skipBuilding = true;
         return true;
     }
+    m_skipBuilding = false;
 
-    QString buildTargetSdk = AndroidManager::buildTargetSDK(target());
-    if (buildTargetSdk.isEmpty()) {
+    if (m_buildTargetSdk.isEmpty()) {
         emit addOutput(tr("Android build SDK not defined. Check Android settings."),
                        OutputFormat::Stderr);
         return false;
     }
 
-    QStringList arguments = {"--input", inputFile,
+    QStringList arguments = {"--input", m_inputFile,
                              "--output", outputDir,
-                             "--android-platform", AndroidManager::buildTargetSDK(target()),
+                             "--android-platform", m_buildTargetSdk,
                              "--jdk", AndroidConfigurations::currentConfig().openJDKLocation().toString()};
 
     if (m_verbose)
@@ -225,6 +254,9 @@ bool AndroidBuildApkStep::init(QList<const BuildStep *> &earlierSteps)
 
     arguments << "--gradle";
 
+    if (m_buildAAB)
+        arguments << "--aab" <<  "--jarsigner";
+
     if (m_useMinistro)
         arguments << "--deployment" << "ministro";
 
@@ -257,7 +289,7 @@ bool AndroidBuildApkStep::init(QList<const BuildStep *> &earlierSteps)
     // Generate arguments with keystore password concealed
     ProjectExplorer::ProcessParameters pp2;
     setupProcessParameters(&pp2, bc, argumentsPasswordConcealed, command);
-    m_command = pp2.effectiveCommand();
+    m_command = pp2.effectiveCommand().toString();
     m_argumentsPasswordConcealed = pp2.prettyArguments();
 
     return true;
@@ -265,7 +297,7 @@ bool AndroidBuildApkStep::init(QList<const BuildStep *> &earlierSteps)
 
 void AndroidBuildApkStep::showInGraphicalShell()
 {
-    Core::FileUtils::showInGraphicalShell(Core::ICore::mainWindow(), m_apkPath);
+    Core::FileUtils::showInGraphicalShell(Core::ICore::mainWindow(), m_packagePath);
 }
 
 ProjectExplorer::BuildStepConfigWidget *AndroidBuildApkStep::createConfigWidget()
@@ -283,8 +315,8 @@ void AndroidBuildApkStep::processFinished(int exitCode, QProcess::ExitStatus sta
 bool AndroidBuildApkStep::verifyKeystorePassword()
 {
     if (!m_keystorePath.exists()) {
-        addOutput(tr("Cannot sign the package. Invalid keystore path (%1).")
-                  .arg(m_keystorePath.toString()), OutputFormat::ErrorMessage);
+        emit addOutput(tr("Cannot sign the package. Invalid keystore path (%1).")
+                           .arg(m_keystorePath.toString()), OutputFormat::ErrorMessage);
         return false;
     }
 
@@ -303,8 +335,8 @@ bool AndroidBuildApkStep::verifyCertificatePassword()
 {
     if (!AndroidManager::checkCertificateExists(m_keystorePath.toString(), m_keystorePasswd,
                                                  m_certificateAlias)) {
-        addOutput(tr("Cannot sign the package. Certificate alias %1 does not exist.")
-                  .arg(m_certificateAlias), OutputFormat::ErrorMessage);
+        emit addOutput(tr("Cannot sign the package. Certificate alias %1 does not exist.")
+                           .arg(m_certificateAlias), OutputFormat::ErrorMessage);
         return false;
     }
 
@@ -324,14 +356,134 @@ bool AndroidBuildApkStep::verifyCertificatePassword()
     return success;
 }
 
-void AndroidBuildApkStep::run(QFutureInterface<bool> &fi)
+
+static bool copyFileIfNewer(const QString &sourceFileName,
+                            const QString &destinationFileName)
+{
+    if (sourceFileName == destinationFileName)
+        return true;
+    if (QFile::exists(destinationFileName)) {
+        QFileInfo destinationFileInfo(destinationFileName);
+        QFileInfo sourceFileInfo(sourceFileName);
+        if (sourceFileInfo.lastModified() <= destinationFileInfo.lastModified())
+            return true;
+        if (!QFile(destinationFileName).remove())
+            return false;
+    }
+
+    if (!QDir().mkpath(QFileInfo(destinationFileName).path()))
+        return false;
+    return QFile::copy(sourceFileName, destinationFileName);
+}
+
+void AndroidBuildApkStep::doRun()
 {
     if (m_skipBuilding) {
-        emit addOutput(tr("No application .pro file found, not building an APK."), BuildStep::OutputFormat::ErrorMessage);
-        reportRunResult(fi, true);
+        emit addOutput(tr("Android deploy settings file not found, not building an APK."), BuildStep::OutputFormat::ErrorMessage);
+        emit finished(true);
         return;
     }
-    AbstractProcessStep::run(fi);
+
+    auto setup = [this] {
+        auto bc = buildConfiguration();
+        const auto androidAbis = AndroidManager::applicationAbis(target());
+        for (const auto &abi : androidAbis) {
+            Utils::FilePath androidLibsDir = bc->buildDirectory()
+                    .pathAppended("android-build/libs")
+                    .pathAppended(abi);
+            if (!androidLibsDir.exists() && !QDir{bc->buildDirectory().toString()}.mkpath(androidLibsDir.toString()))
+                return false;
+        }
+
+        RunConfiguration *rc = target()->activeRunConfiguration();
+        const QString buildKey = rc ? rc->buildKey() : QString();
+        const ProjectNode *node = rc ? target()->project()->findNodeForBuildKey(buildKey) : nullptr;
+
+        if (!node)
+            return false;
+
+        bool inputExists = QFile::exists(m_inputFile);
+        if (inputExists && !AndroidManager::isQtCreatorGenerated(FilePath::fromString(m_inputFile)))
+            return true; // use the generated file if it was not generated by qtcreator
+
+        auto targets = node->data(Android::Constants::AndroidTargets).toStringList();
+        if (targets.isEmpty())
+            return inputExists; // qmake does this job for us
+
+
+        QtSupport::BaseQtVersion *version = QtSupport::QtKitAspect::qtVersion(target()->kit());
+        if (!version)
+            return false;
+
+        QJsonObject deploySettings = Android::AndroidManager::deploymentSettings(target());
+        QString applicationBinary;
+        if (version->qtVersion() < QtSupport::QtVersionNumber(5, 14, 0)) {
+            QTC_ASSERT(androidAbis.size() == 1, return false);
+            applicationBinary = target()->activeRunConfiguration()->buildTargetInfo().targetFilePath.toString();
+            Utils::FilePath androidLibsDir = bc->buildDirectory().pathAppended("android-build/libs").pathAppended(androidAbis.first());
+            for (const auto &target : targets) {
+                if (!copyFileIfNewer(target, androidLibsDir.pathAppended(QFileInfo{target}.fileName()).toString()))
+                    return false;
+            }
+            deploySettings["target-architecture"] = androidAbis.first();
+        } else {
+            applicationBinary = target()->activeRunConfiguration()->buildTargetInfo().targetFilePath.toFileInfo().fileName();
+            QJsonObject architectures;
+
+            // Copy targets to android build folder
+            for (const auto &abi : androidAbis) {
+                QString targetSuffix = QString{"_%1.so"}.arg(abi);
+                if (applicationBinary.endsWith(targetSuffix)) {
+                    // Keep only TargetName from "lib[TargetName]_abi.so"
+                    applicationBinary.remove(0, 3).chop(targetSuffix.size());
+                }
+
+                Utils::FilePath androidLibsDir = bc->buildDirectory()
+                                                     .pathAppended("android-build/libs")
+                                                     .pathAppended(abi);
+                for (const auto &target : targets) {
+                    if (target.endsWith(targetSuffix)) {
+                        if (!copyFileIfNewer(target, androidLibsDir.pathAppended(QFileInfo{target}.fileName()).toString()))
+                            return false;
+                        architectures[abi] = AndroidManager::archTriplet(abi);
+                    }
+                }
+            }
+            deploySettings["architectures"] = architectures;
+        }
+        deploySettings["application-binary"] = applicationBinary;
+
+        QString extraLibs = node->data(Android::Constants::AndroidExtraLibs).toString();
+        if (!extraLibs.isEmpty())
+            deploySettings["android-extra-libs"] = extraLibs;
+
+        QString androidSrcs = node->data(Android::Constants::AndroidPackageSourceDir).toString();
+        if (!androidSrcs.isEmpty())
+            deploySettings["android-package-source-directory"] = androidSrcs;
+
+        QString qmlImportPath = node->data("QML_IMPORT_PATH").toString();
+        if (!qmlImportPath.isEmpty())
+            deploySettings["qml-import-paths"] = qmlImportPath;
+
+        QString qmlRootPath = node->data("QML_ROOT_PATH").toString();
+        if (qmlRootPath.isEmpty())
+            qmlRootPath = target()->project()->rootProjectDirectory().toString();
+         deploySettings["qml-root-path"] = qmlRootPath;
+
+        QFile f{m_inputFile};
+        if (!f.open(QIODevice::WriteOnly))
+            return false;
+        f.write(QJsonDocument{deploySettings}.toJson());
+        return true;
+    };
+
+    if (!setup()) {
+        emit addOutput(tr("Cannot set up Android, not building an APK."), BuildStep::OutputFormat::ErrorMessage);
+        emit finished(false);
+        return;
+    }
+
+    AbstractProcessStep::doRun();
 }
 
 void AndroidBuildApkStep::processStarted()
@@ -344,7 +496,7 @@ void AndroidBuildApkStep::processStarted()
 
 bool AndroidBuildApkStep::fromMap(const QVariantMap &map)
 {
-    m_keystorePath = Utils::FileName::fromString(map.value(KeystoreLocationKey).toString());
+    m_keystorePath = Utils::FilePath::fromString(map.value(KeystoreLocationKey).toString());
     m_signPackage = false; // don't restore this
     m_buildTargetSdk = map.value(BuildTargetSdkKey).toString();
     if (m_buildTargetSdk.isEmpty()) {
@@ -366,7 +518,7 @@ QVariantMap AndroidBuildApkStep::toMap() const
     return map;
 }
 
-Utils::FileName AndroidBuildApkStep::keystorePath()
+Utils::FilePath AndroidBuildApkStep::keystorePath()
 {
     return m_keystorePath;
 }
@@ -379,10 +531,30 @@ QString AndroidBuildApkStep::buildTargetSdk() const
 void AndroidBuildApkStep::setBuildTargetSdk(const QString &sdk)
 {
     m_buildTargetSdk = sdk;
-    AndroidManager::updateGradleProperties(target());
 }
 
-void AndroidBuildApkStep::setKeystorePath(const Utils::FileName &path)
+QVariant AndroidBuildApkStep::data(Core::Id id) const
+{
+    if (id == Constants::AndroidNdkPlatform) {
+        if (auto qtVersion = QtKitAspect::qtVersion(target()->kit()))
+            return AndroidConfigurations::currentConfig()
+                .bestNdkPlatformMatch(AndroidManager::minimumSDK(target()), qtVersion).mid(8);
+        return {};
+    }
+    if (id == Constants::NdkLocation) {
+        if (auto qtVersion = QtKitAspect::qtVersion(target()->kit()))
+            return QVariant::fromValue(AndroidConfigurations::currentConfig().ndkLocation(qtVersion));
+        return {};
+    }
+    if (id == Constants::SdkLocation)
+        return QVariant::fromValue(AndroidConfigurations::currentConfig().sdkLocation());
+    if (id == Constants::AndroidABIs)
+        return AndroidManager::applicationAbis(target());
+
+    return AbstractProcessStep::data(id);
+}
+
+void AndroidBuildApkStep::setKeystorePath(const Utils::FilePath &path)
 {
     m_keystorePath = path;
     m_certificatePasswd.clear();
@@ -414,6 +586,16 @@ void AndroidBuildApkStep::setSignPackage(bool b)
     m_signPackage = b;
 }
 
+bool AndroidBuildApkStep::buildAAB() const
+{
+    return m_buildAAB;
+}
+
+void AndroidBuildApkStep::setBuildAAB(bool aab)
+{
+    m_buildAAB = aab;
+}
+
 bool AndroidBuildApkStep::openPackageLocation() const
 {
     return m_openPackageLocation;
@@ -466,8 +648,8 @@ QAbstractItemModel *AndroidBuildApkStep::keystoreCertificates()
 
     Utils::SynchronousProcess keytoolProc;
     keytoolProc.setTimeoutS(30);
-    const Utils::SynchronousProcessResponse response
-            = keytoolProc.run(AndroidConfigurations::currentConfig().keytoolPath().toString(), params);
+    const SynchronousProcessResponse response
+            = keytoolProc.run({AndroidConfigurations::currentConfig().keytoolPath(), params});
     if (response.result > Utils::SynchronousProcessResponse::FinishedError)
         QMessageBox::critical(nullptr, tr("Error"), tr("Failed to run keytool."));
     else
@@ -486,20 +668,12 @@ PasswordInputDialog::PasswordInputDialog(PasswordInputDialog::Context context,
 {
     inputEdit->setEchoMode(QLineEdit::Password);
 
-    warningIcon->setPixmap(Utils::Icons::WARNING.pixmap());
-    warningIcon->setSizePolicy(QSizePolicy(QSizePolicy::Maximum, QSizePolicy::Maximum));
-    warningIcon->hide();
-
     warningLabel->hide();
 
-    auto warningLayout = new QHBoxLayout;
-    warningLayout->addWidget(warningIcon);
-    warningLayout->addWidget(warningLabel);
-
     auto mainLayout = new QVBoxLayout(this);
     mainLayout->addWidget(inputContextlabel);
     mainLayout->addWidget(inputEdit);
-    mainLayout->addLayout(warningLayout);
+    mainLayout->addWidget(warningLabel);
     mainLayout->addWidget(buttonBox);
 
     connect(inputEdit, &QLineEdit::textChanged,[this](const QString &text) {
@@ -510,9 +684,7 @@ PasswordInputDialog::PasswordInputDialog(PasswordInputDialog::Context context,
         if (verifyCallback(inputEdit->text())) {
             accept(); // Dialog accepted.
         } else {
-            warningIcon->show();
             warningLabel->show();
-            warningLabel->setText(tr("Incorrect password."));
             inputEdit->clear();
             adjustSize();
         }
@@ -551,8 +723,7 @@ namespace Internal {
 
 AndroidBuildApkStepFactory::AndroidBuildApkStepFactory()
 {
-    registerStep<AndroidBuildApkStep>(ANDROID_BUILD_APK_ID);
-    setSupportedProjectType(QmakeProjectManager::Constants::QMAKEPROJECT_ID);
+    registerStep<AndroidBuildApkStep>(Constants::ANDROID_BUILD_APK_ID);
     setSupportedDeviceType(Constants::ANDROID_DEVICE_TYPE);
     setSupportedStepList(ProjectExplorer::Constants::BUILDSTEPS_BUILD);
     setDisplayName(AndroidBuildApkStep::tr("Build Android APK"));
